/*
 * Copyright (c) 2024 ShinGeTsu Meter.
 *
 * SPDX-License-Identifier: BSD-3-Clause
 */
#include <math.h>
#include <stdio.h>
#include <stdlib.h>

#include "hardware/gpio.h"
#include "hardware/watchdog.h"
#include "pico/bootrom.h"
#include "pico/time.h"

#include "i2cDev.h"
#include "meter.h"

// DAC
int16_t dacValue = 0;                   // D/Aコンバータの設定値
int16_t dacRangeMin = DAC_MIN_VALUE;    // TAが0.94となるD/Aコンバータの設定値
int16_t dacRangeMax = DAC_MAX_VALUE;    // TAが6.50となるD/Aコンバータの設定値

// 電圧値
double constV = 0;     // 定電圧
double dacV = 0;       // DAC電圧
double pcV = 0;        // PC電圧
double outputV = 0;    // メータ出力電圧

// TA値
double pcTA = 0;     // PCのTA値
double dacTA = 0;    // DACのTA値

// TAカウンター
double baseTA = 0;       // 集計基準値
double taCounter = 0;    // TAカウンター

// TAの履歴
double   taList[TA_HIST_SIZE];
uint16_t taListIndex = 0;
double   taRange;    // TAの変化量

// イベントフラグ
bool taReset = false;      // リセット
bool armMoved = false;     // トーンアーム操作
bool lcdUpdate = false;    // LCD表示更新

// 状態管理
bool   autoReset = false;    // 自動リセットモード
bool   blocking = false;     // 自動リセット抑止
bool   counting = false;     // TAカウントモード
view_t view = VIEW_1;        // LCD表示モード

/**
 * @brief イベント番号ごとの実行時間を管理し、短時間の連続実行を抑止する。
 */
bool isEventEnabled(uint8_t eventNo) {
    static uint64_t nextAction[] = {0, 0, 0, 0, 0};
    bool            result = false;
    uint64_t        nowUs = time_us_64();
    if (nowUs > nextAction[eventNo]) {
        nextAction[eventNo] = nowUs + EVENT_BLOCK_MS * 1000;
        result = true;
    }
    return result;
}

/**
 * @brief 割り込みハンドラ
 */
void irqHandler(uint gpio, uint32_t events) {
    // リセットボタン
    if (gpio == BUTTON_RST) {
        // AUTO+RESET 表示切替
        if (!gpio_get(BUTTON_AUTO)) {
            if (isEventEnabled(0)) {    // Event0
                view = (++view) % 3;
            }
        }
        // リセット実行
        else {
            if (isEventEnabled(1)) {    // Event1
                taReset = true;
            }
        }
    }

    // カウントモードボタン
    else if (gpio == BUTTON_COUNT) {
        // AUTO+COUNT USB Boot
        if (!gpio_get(BUTTON_AUTO)) {
            reset_usb_boot(0, 0);
            while (1)
                ;
        }
        // RST+COUNT  TAカウンタリセット
        else if (!gpio_get(BUTTON_RST)) {
            if (isEventEnabled(2)) {    // Event2
                baseTA = 0;
                taCounter = 0;
            }
        }
        // カウントモード切替
        else {
            if (isEventEnabled(3)) {    // Event3
                counting = !counting;
                lcdUpdate = true;
            }
        }
    }

    // 自動モードボタン
    else if (gpio == BUTTON_AUTO) {
        // RESET+AUTO  マイコンリセット
        if (!gpio_get(BUTTON_RST)) {
            watchdog_enable(100, true);
            while (1)
                ;
        }
        // 自動モード切替
        else {
            if (isEventEnabled(4)) {    // Event4
                autoReset = !autoReset;
                if (!autoReset) {
                    blocking = false;
                }
                lcdUpdate = true;
            }
        }
    }

    // ロータリーエンコーダー
    else if (gpio == TA_RE1 || gpio == TA_RE2) {
        static const int8_t RE_DIR[] = ROTARY_DIRECTION;

        if (!taReset) {
            static uint8_t taRE1;
            static uint8_t taRE2;
            static size_t  taIndex = 0;
            size_t         lastIndex;
            uint8_t        state = 0;
            if (events == GPIO_IRQ_EDGE_RISE) {
                state = 1;
            }

            if (gpio == TA_RE1) {
                taRE1 = state;
            } else if (gpio == TA_RE2) {
                taRE2 = state;
            }

            lastIndex = taIndex;
            taIndex = ((taIndex << 2) + (taRE1 << 1) + taRE2) & 15;
            if (lastIndex != taIndex) {
                int8_t delta = RE_DIR[taIndex];
                if (delta != 0) {
                    dacValue += delta;
                    armMoved = true;
                }
            }
        }
    }
}

/**
 * @brief GPIO初期化
 */
void initIO() {
    // GPIO初期化
    gpio_init(LED_PIN);
    gpio_set_dir(LED_PIN, GPIO_OUT);

    gpio_init(BUTTON_RST);
    gpio_set_dir(BUTTON_RST, GPIO_IN);
    gpio_pull_up(BUTTON_RST);
    gpio_init(BUTTON_AUTO);
    gpio_set_dir(BUTTON_AUTO, GPIO_IN);
    gpio_pull_up(BUTTON_AUTO);
    gpio_init(BUTTON_COUNT);
    gpio_set_dir(BUTTON_COUNT, GPIO_IN);
    gpio_pull_up(BUTTON_COUNT);

    gpio_init(TA_RE1);
    gpio_set_dir(TA_RE1, GPIO_IN);
    gpio_pull_up(TA_RE1);
    gpio_init(TA_RE2);
    gpio_set_dir(TA_RE2, GPIO_IN);
    gpio_pull_up(TA_RE2);

    gpio_set_irq_enabled_with_callback(BUTTON_RST, GPIO_IRQ_EDGE_FALL, true, irqHandler);
    gpio_set_irq_enabled_with_callback(BUTTON_AUTO, GPIO_IRQ_EDGE_FALL, true, irqHandler);
    gpio_set_irq_enabled_with_callback(BUTTON_COUNT, GPIO_IRQ_EDGE_FALL, true, irqHandler);
    gpio_set_irq_enabled_with_callback(TA_RE1, GPIO_IRQ_EDGE_FALL, true, irqHandler);
    gpio_set_irq_enabled_with_callback(TA_RE1, GPIO_IRQ_EDGE_RISE, true, irqHandler);
    gpio_set_irq_enabled_with_callback(TA_RE2, GPIO_IRQ_EDGE_FALL, true, irqHandler);
    gpio_set_irq_enabled_with_callback(TA_RE2, GPIO_IRQ_EDGE_RISE, true, irqHandler);
}

/**
 * @brief  履歴リストをクリアする
 */
void clearList() {
    for (int i = 0; i < TA_HIST_SIZE; i++) {
        taList[i] = pcTA;
    }
}

/**
 * @brief PCのTAを履歴を保存する。履歴からTAの変動範囲を集計する。
 */
void addTaList(double ta) {
    taList[taListIndex] = ta;

    int    baseIndex = taListIndex + TA_HIST_SIZE;
    double maxValue = 0.0;
    int    maxIndex = 0;
    double minValue = 7.0;
    int    minIndex = 0;
    for (int i = 0; i < TA_HIST_SIZE; i++) {
        int index = (baseIndex - i) % TA_HIST_SIZE;
        if (maxValue < taList[index]) {
            maxValue = taList[index];
            maxIndex = i;
        }
        if (minValue > taList[index]) {
            minValue = taList[index];
            minIndex = i;
        }
    }
    taRange = maxValue - minValue;
    if (minIndex < maxIndex) {
        taRange = -taRange;
    }

    ++taListIndex;
    taListIndex = taListIndex % TA_HIST_SIZE;
}

/**
 * @brief TAアクションをカウントする。
 */
void countTA() {
    if (counting) {
        if (baseTA > dacTA) {
            taCounter += (baseTA - dacTA);
        }
    }
    baseTA = dacTA;
}

/**
 * @brief LCDを更新する。
 */
void update() {
    static char labelCount[6];
    if (counting) {
        sprintf(labelCount, "count");
    } else {
        sprintf(labelCount, "     ");
    }

    static char labelAuto[6];
    if (taReset) {
        sprintf(labelAuto, "reset");
    } else if (blocking) {
        sprintf(labelAuto, "block");
    } else if (autoReset) {
        sprintf(labelAuto, " auto");
    } else {
        sprintf(labelAuto, "     ");
    }

    static char message[30];
    lcdClear();
    if (view == VIEW_1) {    // 通常表示
        sprintf(message, "%4.2f %5.2f %s", pcTA, pcTA - dacTA, labelCount);
        lcdPrint(1, message);
        sprintf(message, "%4.2f %s%6.2f", dacTA, labelAuto, taCounter);
        lcdPrint(2, message);
    } else if (view == VIEW_2) {    // デバッグ用 TA詳細
        sprintf(message, "%6.4f %+7.4f", pcTA, pcTA - dacTA);
        lcdPrint(1, message);
        sprintf(message, "%6.4f%s", dacTA, labelAuto);
        lcdPrint(2, message);
    } else if (view == VIEW_3) {    // デバッグ用 電圧詳細
        sprintf(message, "%6.4f% 6.4f", pcV, outputV);
        lcdPrint(1, message);
        sprintf(message, "%6.4f%s %4d", dacV, labelAuto, dacValue);
        lcdPrint(2, message);
    }
}

/**
 * @brief dacRangeMinとdacRangeMaxを計測する。
 */
double adjustDacRange() {
    gpio_put(LED_PIN, true);
    constV = readConstV();

    // dacRangeMax
    dacValue = setDacValue(DAC_MAX_VALUE);
    dacV = readDacV();
    while ((dacValue > DAC_MIN_VALUE) && (dacV > constV)) {
        dacValue -= 8;
        dacValue = setDacValue(dacValue);
        dacV = readDacV();
    }
    while ((dacValue < DAC_MAX_VALUE) && (dacV < constV)) {
        dacValue += 1;
        dacValue = setDacValue(dacValue);
        dacV = readDacV();
    }
    dacValue -= 1;
    dacRangeMax = setDacValue(dacValue);

    // dacRangeMin
    dacValue = setDacValue(DAC_MIN_VALUE);
    dacV = readDacV();
    while ((dacValue < DAC_MAX_VALUE) && (dacV < 0.0)) {
        dacValue += 512;
        dacValue = setDacValue(dacValue);
        dacV = readDacV();
    }
    while ((dacValue > DAC_MIN_VALUE) && (dacV > 0.0)) {
        dacValue -= 64;
        dacValue = setDacValue(dacValue);
        dacV = readDacV();
    }
    while ((dacValue < DAC_MAX_VALUE) && (dacV < 0.0)) {
        dacValue += 8;
        dacValue = setDacValue(dacValue);
        dacV = readDacV();
    }
    while ((dacValue > DAC_MIN_VALUE) && (dacV > 0.0)) {
        dacValue -= 1;
        dacValue = setDacValue(dacValue);
        dacV = readDacV();
    }
    dacValue += 1;
    dacValue = setDacValue(dacValue);
    dacRangeMin = dacValue;

    dacV = readDacV();
    dacTA = ((dacV / constV) * (TA_MAX - TA_MIN)) + TA_MIN;
    gpio_put(LED_PIN, false);
    update();
}

/**
 * @brief 針をセット位置へ移動する。
 */
void resetTA() {
    // セット位置計算の校正用
    // dacValue = (pcV / constV) * fa + fb;
    static double fa = 0.0;
    static double fb = 0.0;
    // (vRate:pcV/ConstV,value:dacValue)
    static double  p1vRate = 0.0;
    static int16_t p1value = -1.0;
    static double  p2vRate = 1.0;
    static int16_t p2value = DAC_MAX_VALUE;

    bool optimization = false;

    // 初回のみ実行
    if (p1value < 0) {
        p1value = dacRangeMin;
        p2value = dacRangeMax;
        fa = (p1value - p2value) / (p1vRate - p2vRate);
        fb = p1value - p1vRate * (p1value - p2value) / (p1vRate - p2vRate);
    }

    // SET位置へジャンプ
    constV = readConstV();
    dacValue = (pcV / constV) * fa + fb;
    dacValue = setDacValue(dacValue);

    // 微調整
    dacV = readDacV();
    bool dir = pcV > dacV;
    while (taReset) {
        int16_t tValue = dacValue;
        if (dir) {
            dacValue++;
        } else {
            dacValue--;
        }
        dacValue = setDacValue(dacValue);

        if (tValue == dacValue) {
            taReset = false;
            break;
        }

        double tDacV = dacV;
        dacV = readDacV();
        // セット位置を通り過ぎたら終了
        if (dir != (pcV > dacV)) {
            // 通り過ぎる前の方が近かったら戻す
            if (fabs(pcV - tDacV) < fabs(pcV - dacV)) {
                dacValue = setDacValue(tValue);
            }
            optimization = true;
            taReset = false;
        }
    }

    // SET位置計算式を最適化
    if (optimization) {
        // 近いほうを入れ替え
        if (abs(p1value - dacValue) < abs(p2value - dacValue)) {
            // 入れ替え後の差が50以下ならば入れ替えない
            if (abs(p2value - dacValue) > 50) {
                p1vRate = pcV / constV;
                p1value = dacValue;
            }
        } else {
            // 入れ替え後の差が50以下ならば入れ替えない
            if (abs(p1value - dacValue) > 50) {
                p2vRate = pcV / constV;
                p2value = dacValue;
            }
        }
        fa = (p1value - p2value) / (p1vRate - p2vRate);
        fb = p1value - p1vRate * (p1value - p2value) / (p1vRate - p2vRate);
    }

    dacTA = ((dacV / constV) * (TA_MAX - TA_MIN)) + TA_MIN;
}

/**
 * @brief 状態を確認する。
 */
void checkState() {
    if (autoReset) {
        if (blocking) {
            if (fabs(taRange) < TA_BLOCK_TH) {
                blocking = false;
            }
        } else {
            static uint64_t outRangeWait = 0;
            uint64_t        nowUS = time_us_64();
            if ((outputV <= V_METER_MIN) || (V_METER_MAX <= outputV)) {    // メータ範囲外
                if (fabs(taRange) > TA_BLOCK_TH) {                         // ボディモーション
                    blocking = true;
                    counting = false;
                    // ブロック中は振り切ったままとして不規則な動きを抑止
                    if (outputV <= V_METER_MIN) {
                        dacValue = setDacValue(DAC_MIN_VALUE);    
                    } else {
                        dacValue = setDacValue(DAC_MAX_VALUE);
                    }
                } else if (nowUS > outRangeWait) {    // 待ち時間経過
                    taReset = true;
                    outRangeWait = nowUS + RESET_WAIT_MS * 1000;
                }
            } else {
                outRangeWait = nowUS + RESET_WAIT_MS * 1000;
            }
        }
    }
}

/**
 * @brief メイン
 */
int main() {
    sleep_ms(400);
    initIO();
    initI2C();

    adjustDacRange();

    blocking = true;
    autoReset = true;
    taReset = false;
    view = VIEW_1;
    uint64_t nextUpdateUs = 0;
    while (true) {
        while (time_us_64() < nextUpdateUs) {
            if (armMoved || taReset || lcdUpdate) {
                break;
            }
            sleep_us(500);
        }

        pcV = readPcV();
        pcTA = ((pcV / constV) * (TA_MAX - TA_MIN)) + TA_MIN;
        addTaList(pcTA);
        outputV = readOutputV();
        checkState();

        if (armMoved) {
            dacValue = setDacValue(dacValue);
            dacV = readDacV();
            dacTA = ((dacV / constV) * (TA_MAX - TA_MIN)) + TA_MIN;
            countTA();
            armMoved = false;
            update();
        } else if (taReset) {
            update();
            gpio_put(LED_PIN, true);
            resetTA();
            countTA();
            clearList();
            gpio_put(LED_PIN, false);
            update();
        } else if (lcdUpdate) {
            update();
            lcdUpdate = false;
        } else {
            static int32_t count = 1;
            if (count >= UPDATE_RATIO) {
                update();
                count = 1;
            } else {
                count++;
            }
            uint64_t nowUs = time_us_64();
            nextUpdateUs = nextUpdateUs + SAMPLE_WAIT_MS * 1000;
            if (nextUpdateUs < nowUs) {
                nextUpdateUs = nowUs + SAMPLE_WAIT_MS * 1000;
            }
        }
    }
    return 0;
}